import pkg_resources

__all__ = ()


def add_parser(parser, subparser):
    version_parser = subparser.add_parser("version")
    version_parser.set_defaults(func=version)
    return (version_parser, None)


def version() -> str:
    return pkg_resources.get_distribution("netsuite").version



from functools import wraps
from typing import Any, Callable, Optional

from .. import constants
from . import zeep
from .exceptions import NetsuiteResponseError

__all__ = ("WebServiceCall",)


def WebServiceCall(
    path: Optional[str] = None,
    extract: Optional[Callable] = None,
    *,
    default: Any = constants.NOT_SET,
) -> Callable:
    """
    Decorator for NetSuite methods returning SOAP responses

    Args:
        path:
            A dot-separated path for specifying where relevant data resides (where the `status` attribute is set)
        extract:
            A function to extract data from response before returning it.
        default:
            If the existing path does not exist in response, return this
            instead.

    Returns:
        Decorator to use on `NetSuite` web service methods
    """

    def decorator(fn):
        @wraps(fn)
        def wrapper(self, *args, **kw):
            response = fn(self, *args, **kw)
            if not isinstance(response, zeep.xsd.ComplexType):
                return response

            if path is not None:
                for part in path.split("."):
                    try:
                        response = getattr(response, part)
                    except AttributeError:
                        if default is constants.NOT_SET:
                            raise
                        else:
                            return default

            try:
                response_status = response["status"]
            except TypeError:
                response_status = None
                for record in response:
                    # NOTE: Status is set on each returned record for lists,
                    #       really strange...
                    response_status = record["status"]
                    break

            is_success = response_status["isSuccess"]

            if not is_success:
                response_detail = response_status["statusDetail"]
                raise NetsuiteResponseError(response_detail)

            if extract is not None:
                response = extract(response)

            return response

        return wrapper

    return decorator



class NetsuiteAPIRequestError(Exception):
    """Raised when a Netsuite REST API request fails"""

    def __init__(self, status_code: int, response_text: str):
        self.status_code = status_code
        self.response_text = response_text

    def __str__(self):
        return f"HTTP{self.status_code} - {self.response_text}"


class NetsuiteAPIResponseParsingError(NetsuiteAPIRequestError):
    """Raised when parsing a Netsuite REST API response fails"""



import logging
from functools import cached_property
from typing import Sequence

from . import rest_api_base
from .config import Config

logger = logging.getLogger(__name__)

__all__ = ("NetSuiteRestApi",)


class NetSuiteRestApi(rest_api_base.RestApiBase):
    def __init__(
        self,
        config: Config,
        *,
        default_timeout: int = 60,
        concurrent_requests: int = 10,
        signature_method: str = rest_api_base.DEFAULT_SIGNATURE_METHOD,
    ):
        self._config = config
        self._default_timeout = default_timeout
        self._concurrent_requests = concurrent_requests
        self._signature_method = signature_method

    @cached_property
    def hostname(self) -> str:
        return self._make_hostname()

    async def get(self, subpath: str, **request_kw):
        return await self._request("GET", subpath, **request_kw)

    async def post(self, subpath: str, **request_kw):
        return await self._request(
            "POST",
            subpath,
            **request_kw,
        )

    async def put(self, subpath: str, **request_kw):
        return await self._request("PUT", subpath, **request_kw)

    async def patch(self, subpath: str, **request_kw):
        return await self._request("PATCH", subpath, **request_kw)

    async def delete(self, subpath: str, **request_kw):
        return await self._request("DELETE", subpath, **request_kw)

    async def suiteql(self, q: str, limit: int = 10, offset: int = 0, **request_kw):
        """
        Example:
        >>> suiteql(q="SELECT * FROM Transaction", limit=10, offset=0)
        """
        return await self._request(
            "POST",
            # https://docs.oracle.com/en/cloud/saas/netsuite/ns-online-help/section_156257799794.html#Using-SuiteQL
            "/query/v1/suiteql",
            headers={"Prefer": "transient", **request_kw.pop("headers", {})},
            json={"q": q, **request_kw.pop("json", {})},
            params={"limit": limit, "offset": offset, **request_kw.pop("params", {})},
            **request_kw,
        )

    async def jsonschema(self, record_type: str, **request_kw):
        headers = {
            "Accept": "application/schema+json",
            **request_kw.pop("headers", {}),
        }
        return await self._request(
            "GET",
            f"/record/v1/metadata-catalog/{record_type}",
            headers=headers,
            **request_kw,
        )

    async def token_info(self, **request_kw):
        """
        Retrieves metadata about the current token. Role, company, etc.

        https://docs.oracle.com/en/cloud/saas/netsuite/ns-online-help/chapter_157017286140.html#Issue-Token-and-Revoke-Token-REST-Services-for-Token-based-Authentication
        """

        # this overrides the default URL generation: this specific endpoint hits a completely different host
        request_kw[
            "url"
        ] = f"https://{self._config.account_slugified}.restlets.api.netsuite.com/rest/tokeninfo"

        return await self._request(
            method="GET",
            # useless, but required by _request
            subpath="ignored",
            **request_kw,
        )

    async def openapi(self, record_types: Sequence[str] = (), **request_kw):
        """
        Retrieves the OpenAPI specification (metadata catalog) for the Netsuite REST API. This is the best way to
        introspect the NetSuite account and return the record structure.

        Args:
            record_types (Sequence[str]): Optional. List of record types to include in the OpenAPI specification.
            **request_kw: Optional keyword arguments to be passed to the underlying request.

        Returns:
            The OpenAPI specification as a JSON object.
        """

        headers = {
            "Accept": "application/swagger+json",
            **request_kw.pop("headers", {}),
        }
        params = request_kw.pop("params", {})

        if len(record_types) > 0:
            params["select"] = ",".join(record_types)

        return await self._request(
            "GET",
            "/record/v1/metadata-catalog",
            headers=headers,
            params=params,
            **request_kw,
        )

    def _make_hostname(self):
        return f"{self._config.account_slugified}.suitetalk.api.netsuite.com"

    def _make_url(self, subpath: str):
        return f"https://{self.hostname}/services/rest{subpath}"

    def _make_default_headers(self):
        return {
            "Content-Type": "application/json",
            "X-NetSuite-PropertyNameValidation": "error",
        }



import urllib.parse

from . import zeep

__all__ = ("AsyncNetSuiteTransport",)


# TODO: ASYNC! Maybe remove this custom transport?!?!


class AsyncNetSuiteTransport(zeep.transports.AsyncTransport):
    """
    NetSuite company-specific domain wrapper for zeep.transports.transport

    Latest NetSuite WSDL now uses relative definition addresses

    zeep maps reflective remote calls to the base WSDL address,
    rather than the dynamic subscriber domain

    Wrap the zeep transports service with our address modifications
    """

    def __init__(self, wsdl_url, *args, **kwargs):
        parsed = urllib.parse.urlparse(wsdl_url)
        self._netsuite_base_url = f"{parsed.scheme}://{parsed.netloc}"
        super().__init__(*args, **kwargs)

    def _fix_address(self, address):
        """Munge the address to the company-specific domain, not the default"""
        idx = address.index("/", 8)
        path = address[idx:]
        return f"{self._netsuite_base_url}{path}"

    async def get(self, address, params, headers):
        return await super().get(self._fix_address(address), params, headers)

    async def post(self, address, message, headers):
        return await super().post(self._fix_address(address), message, headers)



import argparse
import functools
import http.server
import logging
import logging.config
import pathlib
import tempfile
from typing import Dict, List, Optional, Union

from .. import json
from ..client import NetSuite
from ..config import Config

logger = logging.getLogger("netsuite")

ParsedHeaders = Dict[str, Union[List[str], str]]

__all__ = ()


def add_parser(parser, subparser):
    rest_api_parser = subparser.add_parser("rest-api")
    rest_api_subparser = rest_api_parser.add_subparsers()
    _add_rest_api_get_parser(rest_api_parser, rest_api_subparser)
    _add_rest_api_post_parser(rest_api_parser, rest_api_subparser)
    _add_rest_api_put_parser(rest_api_parser, rest_api_subparser)
    _add_rest_api_patch_parser(rest_api_parser, rest_api_subparser)
    _add_rest_api_delete_parser(rest_api_parser, rest_api_subparser)
    _add_rest_api_suiteql_parser(rest_api_parser, rest_api_subparser)
    _add_rest_api_jsonschema_parser(rest_api_parser, rest_api_subparser)
    _add_rest_api_openapi_parser(rest_api_parser, rest_api_subparser)
    _add_rest_api_openapi_serve_parser(rest_api_parser, rest_api_subparser)

    return (rest_api_parser, rest_api_subparser)


def _add_rest_api_get_parser(parser, subparser):
    async def rest_api_get(config, args) -> str:
        rest_api = _get_rest_api_or_error(parser, config)
        params = {}
        if args.expandSubResources is True:
            params["expandSubResources"] = "true"
        if args.limit is not None:
            params["limit"] = args.limit
        if args.offset is not None:
            params["offset"] = args.offset
        if args.fields is not None:
            params["fields"] = ",".join(args.fields)
        if args.expand is not None:
            params["expand"] = ",".join(args.expand)
        if args.query is not None:
            params["q"] = args.query
        resp = await rest_api.get(
            args.subpath, params=params, headers=_parse_headers_arg(parser, args.header)
        )
        return json.dumps(resp)

    p = subparser.add_parser(
        "get", description="Make a GET request to NetSuite REST web services"
    )
    p.set_defaults(func=rest_api_get)
    p.add_argument(
        "subpath",
        help="The subpath to GET, e.g. `/record/v1/salesOrder`",
    )
    p.add_argument(
        "-q",
        "--query",
        help="Search query used to filter results. See NetSuite help center for syntax information. Only works for list endpoints e.g. /record/v1/customer",
    )
    p.add_argument(
        "-e",
        "--expandSubResources",
        action="store_true",
        help="Automatically expand all sublists, sublist lines and subrecords on this record. Only works for detail endpoints e.g. /record/v1/invoice/123.",
    )
    p.add_argument("-l", "--limit", type=int)
    p.add_argument("-o", "--offset", type=int)
    p.add_argument(
        "-f",
        "--fields",
        metavar="field",
        nargs="*",
        help="Only include the given fields in response",
    )
    p.add_argument(
        "-E",
        "--expand",
        nargs="*",
        help="Expand the given sublist lines and subrecords on this record. Only works for detail endpoints e.g. /record/v1/invoice/123.",
    )
    _add_rest_api_headers_arg(p)


def _add_rest_api_post_parser(parser, subparser):
    async def rest_api_post(config, args) -> str:
        rest_api = _get_rest_api_or_error(parser, config)
        with args.payload_file as fh:
            payload_str = fh.read()

        payload = json.loads(payload_str)

        resp = await rest_api.post(
            args.subpath, json=payload, headers=_parse_headers_arg(parser, args.header)
        )
        return json.dumps(resp)

    p = subparser.add_parser(
        "post", description="Make a POST request to NetSuite REST web services"
    )
    p.set_defaults(func=rest_api_post)
    p.add_argument(
        "subpath",
        help="The subpath to POST to, e.g. `/record/v1/salesOrder`",
    )
    p.add_argument("payload_file", type=argparse.FileType("r"))
    _add_rest_api_headers_arg(p)


def _add_rest_api_put_parser(parser, subparser):
    async def rest_api_put(config, args) -> str:
        rest_api = _get_rest_api_or_error(parser, config)
        with args.payload_file as fh:
            payload_str = fh.read()

        payload = json.loads(payload_str)

        resp = await rest_api.put(
            args.subpath, json=payload, headers=_parse_headers_arg(parser, args.header)
        )
        return json.dumps(resp)

    p = subparser.add_parser(
        "put", description="Make a PUT request to NetSuite REST web services"
    )
    p.set_defaults(func=rest_api_put)
    p.add_argument(
        "subpath",
        help="The subpath to PUT to, e.g. `/record/v1/salesOrder/eid:abc123`",
    )
    p.add_argument("payload_file", type=argparse.FileType("r"))
    _add_rest_api_headers_arg(p)


def _add_rest_api_patch_parser(parser, subparser):
    async def rest_api_patch(config, args) -> str:
        rest_api = _get_rest_api_or_error(parser, config)
        with args.payload_file as fh:
            payload_str = fh.read()

        payload = json.loads(payload_str)

        resp = await rest_api.patch(
            args.subpath, json=payload, headers=_parse_headers_arg(parser, args.header)
        )
        return json.dumps(resp)

    p = subparser.add_parser(
        "patch", description="Make a PATCH request to NetSuite REST web services"
    )
    p.set_defaults(func=rest_api_patch)
    p.add_argument(
        "subpath",
        help="The subpath to PATCH to, e.g. `/record/v1/salesOrder/eid:abc123`",
    )
    p.add_argument("payload_file", type=argparse.FileType("r"))
    _add_rest_api_headers_arg(p)


def _add_rest_api_delete_parser(parser, subparser):
    async def rest_api_delete(config, args) -> str:
        rest_api = _get_rest_api_or_error(parser, config)
        resp = await rest_api.delete(
            args.subpath, headers=_parse_headers_arg(parser, args.header)
        )
        return json.dumps(resp)

    p = subparser.add_parser(
        "delete", description="Make a DELETE request to NetSuite REST web services"
    )
    p.set_defaults(func=rest_api_delete)
    p.add_argument(
        "subpath",
        help="The subpath for the DELETE request, e.g. `/record/v1/salesOrder/eid:abc123`",
    )
    _add_rest_api_headers_arg(p)


def _add_rest_api_suiteql_parser(parser, subparser):
    async def rest_api_suiteql(config, args) -> str:
        rest_api = _get_rest_api_or_error(parser, config)

        with args.q_file as fh:
            q = fh.read()

        resp = await rest_api.suiteql(
            q=q,
            limit=args.limit,
            offset=args.offset,
            headers=_parse_headers_arg(parser, args.header),
        )

        return json.dumps(resp)

    p = subparser.add_parser(
        "suiteql", description="Make a SuiteQL request to NetSuite REST web services"
    )
    p.set_defaults(func=rest_api_suiteql)
    p.add_argument(
        "q_file", type=argparse.FileType("r"), help="File containing a SuiteQL query"
    )
    p.add_argument("-l", "--limit", type=int, default=10)
    p.add_argument("-o", "--offset", type=int, default=0)
    _add_rest_api_headers_arg(p)


def _add_rest_api_jsonschema_parser(parser, subparser):
    async def rest_api_jsonschema(config, args) -> str:
        rest_api = _get_rest_api_or_error(parser, config)
        resp = await rest_api.jsonschema(args.record_type)
        return json.dumps(resp)

    p = subparser.add_parser(
        "jsonschema", description="Retrieve JSON Schema for the given record type"
    )
    p.set_defaults(func=rest_api_jsonschema)
    p.add_argument("record_type", help="The record type to get JSONSchema spec for")
    _add_rest_api_headers_arg(p)


def _add_rest_api_openapi_parser(parser, subparser):
    async def rest_api_openapi(config, args) -> str:
        rest_api = _get_rest_api_or_error(parser, config)
        resp = await rest_api.openapi(args.record_types)
        return json.dumps(resp)

    p = subparser.add_parser(
        "openapi",
        description="Retrieve OpenAPI spec for the given record types",
    )
    p.set_defaults(func=rest_api_openapi)
    p.add_argument(
        "record_types",
        metavar="record_type",
        nargs="+",
        help="The record type(s) to get OpenAPI spec for",
    )
    _add_rest_api_headers_arg(p)


def _add_rest_api_openapi_serve_parser(parser, subparser):
    async def rest_api_openapi_serve(config, args):
        rest_api = _get_rest_api_or_error(parser, config)
        if len(args.record_types) == 0:
            logger.warning(
                "Fetching OpenAPI spec for ALL known record types... This will take a long "
                "time! (Consider providing only the record types of interest by passing "
                "their names to this command as positional arguments)"
            )
        else:
            rt_str = ", ".join(args.record_types)
            logger.info(f"Fetching OpenAPI spec for record types {rt_str}...")
        spec = await rest_api.openapi(args.record_types)
        tempdir = pathlib.Path(tempfile.mkdtemp())
        openapi_file = tempdir / "openapi.json"
        html_file = tempdir / "index.html"
        openapi_file.write_bytes(json.dumps(spec).encode("utf-8"))
        html = """<!DOCTYPE html>
    <html>
        <head>
        <link type="text/css" rel="stylesheet" href="https://cdn.jsdelivr.net/npm/swagger-ui-dist@5/swagger-ui.css">
        <title>NetSuite REST Record API</title>
        </head>
        <body>
        <div id="swagger-ui">
        </div>
        <script src="https://cdn.jsdelivr.net/npm/swagger-ui-dist@5/swagger-ui-bundle.js"></script>
        <!-- `SwaggerUIBundle` is now available on the page -->
        <script>
        const ui = SwaggerUIBundle({
            url: '/openapi.json',
            dom_id: '#swagger-ui',
            presets: [
                SwaggerUIBundle.presets.apis,
                SwaggerUIBundle.SwaggerUIStandalonePreset
            ],
            layout: "BaseLayout",
            deepLinking: true
        })
        </script>
        </body>
    </html>"""
        html_file.write_text(html)
        handler_class = functools.partial(
            http.server.SimpleHTTPRequestHandler,
            directory=str(tempdir),
        )
        logger.info(
            f"NetSuite REST Record API docs available at http://{args.bind}:{args.port}"
        )
        try:
            http.server.test(
                HandlerClass=handler_class,
                ServerClass=http.server.ThreadingHTTPServer,
                port=args.port,
                bind=args.bind,
            )
        finally:
            html_file.unlink()
            openapi_file.unlink()
            tempdir.rmdir()

    p = subparser.add_parser(
        "openapi-serve",
        description="Start a HTTP server on localhost serving the OpenAPI spec via Swagger UI",
    )
    p.set_defaults(func=rest_api_openapi_serve)
    p.add_argument(
        "record_types",
        metavar="record_type",
        nargs="*",
        help="The record type(s) to get OpenAPI spec for. If not provided the OpenAPI spec for all known record types will be retrieved.",
    )
    p.add_argument("-p", "--port", default=8000, type=int, help="The port to listen to")
    p.add_argument("-b", "--bind", default="127.0.0.1", help="The host to bind to")


def _get_rest_api_or_error(parser, config: Config):
    ns = NetSuite(config)

    try:
        return ns.rest_api  # Cached property that initializes NetSuiteRestApi
    except RuntimeError as ex:
        parser.error(str(ex))


def _parse_headers_arg(
    parser,
    headers: Optional[List[str]],
) -> ParsedHeaders:
    out: ParsedHeaders = {}

    if headers is None:
        headers = []

    for raw_header in headers:
        err = False
        try:
            k, v = raw_header.split(":", maxsplit=1)
        except ValueError:
            err = True
        else:
            k, v = (k.strip(), v.strip())
            if not k or not v:
                err = True
        if err:
            parser.error(
                f"Invalid header: `{raw_header}``. Should have format: `NAME: VALUE`"
            )
        else:
            existing = out.get(k)
            if existing:
                if isinstance(existing, list):
                    existing.append(v)
                else:
                    out[k] = [existing, v]
            else:
                out[k] = v
    return out


def _add_rest_api_headers_arg(parser):
    parser.add_argument(
        "-H",
        "--header",
        action="append",
        help="Headers to append. Can be specified multiple time and the format for each is `KEY: VALUE`",
    )



import argparse

from ..config import Config

__all__ = ()


def load_config_or_error(parser: argparse.ArgumentParser, path: str, section: str) -> Config:  # type: ignore[return]
    try:
        conf = Config.from_ini(path=path, section=section)
    except FileNotFoundError:
        parser.error(f"Config file {path} not found")
    except KeyError as ex:
        if ex.args == (section,):
            parser.error(f"No config section `{section}` in file {path}")
        else:
            raise ex
    else:
        return conf



import argparse

from .. import json
from ..client import NetSuite
from ..config import Config

__all__ = ()


def add_parser(parser, subparser):
    restlet_parser = subparser.add_parser(
        "restlet", description="Make NetSuite Restlet requests"
    )
    restlet_subparser = restlet_parser.add_subparsers()
    _add_restlet_get_parser(restlet_parser, restlet_subparser)
    _add_restlet_post_parser(restlet_parser, restlet_subparser)
    _add_restlet_put_parser(restlet_parser, restlet_subparser)
    _add_restlet_delete_parser(restlet_parser, restlet_subparser)

    return (restlet_parser, restlet_subparser)


def _add_restlet_get_parser(parser, subparser):
    async def restlet_get(config, args) -> str:
        restlet = _get_restlet_or_error(parser, config)

        resp = await restlet.get(script_id=args.script_id, deploy=args.deploy)
        return json.dumps(resp)

    p = subparser.add_parser(
        "get", description="Make a GET request to NetSuite Restlet"
    )
    _add_default_restlet_args(p)
    p.set_defaults(func=restlet_get)


def _add_restlet_post_parser(parser, subparser):
    async def restlet_post(config, args) -> str:
        restlet = _get_restlet_or_error(parser, config)

        with args.payload_file as fh:
            payload_str = fh.read()

        payload = json.loads(payload_str)

        resp = await restlet.post(
            script_id=args.script_id, deploy=args.deploy, json=payload
        )
        return json.dumps(resp)

    p = subparser.add_parser(
        "post", description="Make a POST request to NetSuite Restlet"
    )
    p.set_defaults(func=restlet_post)
    _add_default_restlet_args(p)
    p.add_argument("payload_file", type=argparse.FileType("r"))


def _add_restlet_put_parser(parser, subparser):
    async def restlet_put(config, args) -> str:
        restlet = _get_restlet_or_error(parser, config)

        with args.payload_file as fh:
            payload_str = fh.read()

        payload = json.loads(payload_str)

        resp = await restlet.put(
            script_id=args.script_id, deploy=args.deploy, json=payload
        )
        return json.dumps(resp)

    p = subparser.add_parser(
        "put", description="Make a PUT request to NetSuite Restlet"
    )
    p.set_defaults(func=restlet_put)
    _add_default_restlet_args(p)
    p.add_argument("payload_file", type=argparse.FileType("r"))


def _add_restlet_delete_parser(parser, subparser):
    async def restlet_delete(config, args) -> str:
        restlet = _get_restlet_or_error(parser, config)

        resp = await restlet.put(script_id=args.script_id, deploy=args.deploy)
        return json.dumps(resp)

    p = subparser.add_parser(
        "delete", description="Make a DELETE request to a NetSuite Restlet"
    )
    p.set_defaults(func=restlet_delete)
    _add_default_restlet_args(p)


def _get_restlet_or_error(parser, config: Config):
    ns = NetSuite(config)

    try:
        return ns.restlet  # Cached property that initializes NetSuiteRestlet
    except RuntimeError as ex:
        parser.error(str(ex))


def _add_default_restlet_args(parser_: argparse.ArgumentParser):
    parser_.add_argument("script_id", type=int, help="The script to run")
    parser_.add_argument(
        "-d", "--deploy", type=int, default=1, help="The deployment version"
    )



import configparser
import os
import typing as t

from pydantic import BaseModel

from .constants import DEFAULT_INI_PATH, DEFAULT_INI_SECTION

__all__ = ("Config", "TokenAuth")


class TokenAuth(BaseModel):
    consumer_key: str
    consumer_secret: str
    token_id: str
    token_secret: str


class Config(BaseModel):
    account: str
    auth: TokenAuth
    log_level: t.Optional[str] = None

    # TODO: Support OAuth2
    # auth: Union[OAuth2, TokenAuth]

    @property
    def account_slugified(self) -> str:
        # https://followingnetsuite.wordpress.com/2018/10/18/suitetalk-sandbox-urls-addendum/
        return self.account.lower().replace("_", "-")

    @staticmethod
    def _reorganize_auth_keys(raw: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
        # we intentionally do not type `reorganized` here, mypy does not like union types with dict key assignment
        # https://stackoverflow.com/questions/69824126/mypy-invalid-index-type-str-for-unionstr-dictstr-str-expected-type-u

        reorganized: t.Dict[t.Any, t.Any] = {"auth": {}}

        for key, val in raw.items():
            if key in TokenAuth.model_fields:
                reorganized["auth"][key] = val
            else:
                reorganized[key] = val
        return reorganized

    @classmethod
    def from_env(cls):
        """
        Initialize config from environment variables.

        - `NETSUITE_AUTH_TYPE`: Specifies the type of authentication, defaults to `token`
        - `NETSUITE_ACCOUNT`: The Netsuite account number.
        - `NETSUITE_CONSUMER_KEY`: The consumer key for OAuth.
        - `NETSUITE_CONSUMER_SECRET`: The consumer secret for OAuth.
        - `NETSUITE_TOKEN_ID`: The token ID for OAuth.
        - `NETSUITE_TOKEN_SECRET`: The token secret for OAuth.
        - `NETSUITE_LOG_LEVEL`: log level for NetSuite debugging

        Returns a dictionary of available config options.
        """

        keys = [
            "auth_type",
            "account",
            "consumer_key",
            "consumer_secret",
            "token_id",
            "token_secret",
            "log_level",
        ]
        prefix = "NETSUITE_"
        raw = {
            k: os.environ[prefix + k.upper()]
            for k in keys
            if prefix + k.upper() in os.environ
        }

        reorganized = cls._reorganize_auth_keys(raw)
        return cls(**reorganized)

    @classmethod
    def from_ini(
        cls, path: str = DEFAULT_INI_PATH, section: str = DEFAULT_INI_SECTION
    ) -> "Config":
        iniconf = configparser.ConfigParser()
        with open(path) as fp:
            iniconf.read_file(fp)

        selected_configuration = iniconf[section]

        auth_type = selected_configuration.get("auth_type", "token")
        if auth_type != "token":
            raise RuntimeError(f"Only token auth is supported, not `{auth_type}`")

        raw = {key: val for key, val in selected_configuration.items()}
        reorganized = cls._reorganize_auth_keys(raw)
        return cls(**reorganized)  # type: ignore



from . import zeep


def to_builtin(obj, *, target_cls=dict):
    """
    Turn zeep XML object into python built-in data structures

    Args:
        target_cls (Type[dict]):
            Which type of dictionary type will be used for objects.
            As this project's minimum Python version officially supported is
            3.6 we can rely on the native sorted order of the standard `dict`
            class as a default.
    """
    return zeep.helpers.serialize_object(obj, target_cls=target_cls)



# Compatibility module - in case SOAP isn't enabled in library
try:
    import zeep as __zeep  # noqa
except ImportError:
    ZEEP_INSTALLED = False
else:
    ZEEP_INSTALLED = True

if ZEEP_INSTALLED:
    import requests
    from zeep import *  # noqa
    from zeep import cache, client, helpers, transports, xsd
else:

    class _Transport:
        ...

    class _BaseCache:
        ...

    class _SqliteCache:
        ...

    class _CompoundValue:
        ...

    class _Client:
        ...

    class _ServiceProxy:
        ...

    class _Factory:
        ...

    class _valueobjects:
        CompoundValue = _CompoundValue

    class cache:  # type: ignore[no-redef]
        Base = _BaseCache
        SqliteCache = _SqliteCache

    class client:  # type: ignore[no-redef]
        Client = _Client
        AsyncClient = _Client
        ServiceProxy = _ServiceProxy
        Factory = _Factory

    class transports:  # type: ignore[no-redef]
        Transport = _Transport
        AsyncTransport = _Transport

    class xsd:  # type: ignore[no-redef]
        CompoundValue = _CompoundValue
        valueobjects = _valueobjects

    class helpers:  # type: ignore[no-redef]
        serialize_object = None

    class requests:  # type: ignore[no-redef]
        Session = None



import datetime
from decimal import Decimal
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, Type, Union
from uuid import UUID

__all__ = ("dumps", "loads")

try:
    import orjson as _json
except ImportError:
    import json as _json  # type: ignore[no-redef]

    HAS_ORJSON = False
else:
    HAS_ORJSON = True


loads = _json.loads


def dumps(obj: Any, *args, **kw) -> str:
    if HAS_ORJSON:
        kw["default"] = _orjson_default
        return _json.dumps(obj, *args, **kw).decode("utf-8")
    else:
        return _json.dumps(obj, *args, **kw)  # type: ignore[return-value]


def _orjson_default(obj: Any) -> Any:
    """Handle cases which orjson doesn't know what to do with"""

    # Handle that orjson doesn't support subclasses of str
    if isinstance(obj, str):
        return str(obj)
    else:
        encoder = _get_encoder(obj)
        return encoder(obj)


def _isoformat(o: Union[datetime.date, datetime.time]) -> str:
    return o.isoformat()


def _get_encoder(obj: Any) -> Any:
    for base in obj.__class__.__mro__[:-1]:
        try:
            encoder = _ENCODERS_BY_TYPE[base]
        except KeyError:
            continue
        return encoder(obj)
    else:  # We have exited the for loop without finding a suitable encoder
        raise TypeError(
            f"Object of type '{obj.__class__.__name__}' is not JSON serializable"
        )


_ENCODERS_BY_TYPE: Dict[Type[Any], Callable[[Any], Any]] = {
    bytes: lambda o: o.decode(),
    datetime.date: _isoformat,
    datetime.datetime: _isoformat,
    datetime.time: _isoformat,
    datetime.timedelta: lambda td: td.total_seconds(),
    Decimal: float,
    Enum: lambda o: o.value,
    frozenset: list,
    Path: str,
    set: list,
    UUID: str,
}



import logging
import re
from contextlib import contextmanager
from datetime import datetime
from functools import cached_property
from typing import Dict, List, Optional, Sequence

from ..config import Config
from . import helpers, passport, zeep
from .decorators import WebServiceCall
from .transports import AsyncNetSuiteTransport

logger = logging.getLogger(__name__)

__all__ = ("NetSuiteSoapApi",)


class NetSuiteSoapApi:
    version = "2021.1.0"
    wsdl_url_tmpl = "https://{account_slug}.suitetalk.api.netsuite.com/wsdl/v{underscored_version}/netsuite.wsdl"

    def __init__(
        self,
        config: Config,
        *,
        version: Optional[str] = None,
        wsdl_url: Optional[str] = None,
        cache: Optional[zeep.cache.Base] = None,
    ) -> None:
        self._ensure_required_dependencies()
        if version is not None:
            assert re.match(r"\d+\.\d+\.\d+", version)
            self.version = version
        self.config: Config = config
        self._wsdl_url: Optional[str] = wsdl_url
        self._cache: Optional[zeep.cache.Base] = cache
        self._client: Optional[zeep.client.AsyncClient] = None

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.hostname}({self.version})>"

    async def __aenter__(self):
        await self.client.__aenter__()
        return self

    async def __aexit__(self, exc_type=None, exc_value=None, traceback=None) -> None:
        await self.client.__aexit__(
            exc_type=exc_type, exc_value=exc_value, traceback=traceback
        )
        # Connection is now closed by zeep. Generate a new one
        self._client = self._generate_client()

    @property
    def wsdl_url(self) -> str:
        if self._wsdl_url is None:
            self._wsdl_url = self._generate_wsdl_url()
        return self._wsdl_url

    @property
    def cache(self) -> zeep.cache.Base:
        if self._cache is None:
            self._cache = self._generate_cache()
        return self._cache

    @property
    def client(self) -> zeep.client.AsyncClient:
        if self._client is None:
            self._client = self._generate_client()
        return self._client

    @property
    def transport(self):
        return self.client.transport

    @cached_property
    def hostname(self) -> str:
        return self.wsdl_url.replace("https://", "").partition("/")[0]

    @property
    def service(self) -> zeep.client.ServiceProxy:
        return self.client.service

    @property
    def underscored_version(self) -> str:
        return self.version.replace(".", "_")

    @property
    def underscored_version_no_micro(self) -> str:
        return self.underscored_version.rpartition("_")[0]

    def _generate_wsdl_url(self) -> str:
        return self.wsdl_url_tmpl.format(
            underscored_version=self.underscored_version,
            account_slug=self.config.account_slugified,
        )

    def _generate_cache(self) -> zeep.cache.Base:
        return zeep.cache.SqliteCache(timeout=60 * 60 * 24 * 365)

    def _generate_transport(self) -> zeep.transports.AsyncTransport:
        return AsyncNetSuiteTransport(
            self.wsdl_url,
            cache=self.cache,
        )

    def generate_passport(self) -> Dict:
        return passport.make(self, self.config)

    def to_builtin(self, obj, *args, **kw):
        """Turn zeep XML object into python built-in data structures"""
        return helpers.to_builtin(obj, *args, **kw)

    @contextmanager
    def with_timeout(self, timeout: int):
        """Run SuiteTalk operation with the specified timeout"""
        with self.transport.settings(timeout=timeout):
            yield

    def _generate_client(self) -> zeep.client.AsyncClient:
        return zeep.client.AsyncClient(
            self.wsdl_url,
            transport=self._generate_transport(),
        )

    def _get_namespace(self, name: str, sub_namespace: str) -> str:
        return "urn:{name}_{version}.{sub_namespace}.webservices.netsuite.com".format(
            name=name,
            version=self.underscored_version_no_micro,
            sub_namespace=sub_namespace,
        )

    def _type_factory(self, name: str, sub_namespace: str) -> zeep.client.Factory:
        return self.client.type_factory(self._get_namespace(name, sub_namespace))

    @classmethod
    def _ensure_required_dependencies(cls):
        if not cls._has_required_dependencies():
            raise RuntimeError(
                "Missing required dependencies for SOAP Web Services API support. "
                "Install with `pip install netsuite[soap_api]`"
            )

    @classmethod
    def _has_required_dependencies(cls) -> bool:
        return zeep.ZEEP_INSTALLED

    @cached_property
    def Core(self) -> zeep.client.Factory:
        return self._type_factory("core", "platform")

    @cached_property
    def CoreTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.core", "platform")

    @cached_property
    def FaultsTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.faults", "platform")

    @cached_property
    def Faults(self) -> zeep.client.Factory:
        return self._type_factory("faults", "platform")

    @cached_property
    def Messages(self) -> zeep.client.Factory:
        return self._type_factory("messages", "platform")

    @cached_property
    def Common(self) -> zeep.client.Factory:
        return self._type_factory("common", "platform")

    @cached_property
    def CommonTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.common", "platform")

    @cached_property
    def Scheduling(self) -> zeep.client.Factory:
        return self._type_factory("scheduling", "activities")

    @cached_property
    def SchedulingTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.scheduling", "activities")

    @cached_property
    def Communication(self) -> zeep.client.Factory:
        return self._type_factory("communication", "general")

    @cached_property
    def CommunicationTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.communication", "general")

    @cached_property
    def Filecabinet(self) -> zeep.client.Factory:
        return self._type_factory("filecabinet", "documents")

    @cached_property
    def FilecabinetTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.filecabinet", "documents")

    @cached_property
    def Relationships(self) -> zeep.client.Factory:
        return self._type_factory("relationships", "lists")

    @cached_property
    def RelationshipsTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.relationships", "lists")

    @cached_property
    def Support(self) -> zeep.client.Factory:
        return self._type_factory("support", "lists")

    @cached_property
    def SupportTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.support", "lists")

    @cached_property
    def Accounting(self) -> zeep.client.Factory:
        return self._type_factory("accounting", "lists")

    @cached_property
    def AccountingTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.accounting", "lists")

    @cached_property
    def Sales(self) -> zeep.client.Factory:
        return self._type_factory("sales", "transactions")

    @cached_property
    def SalesTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.sales", "transactions")

    @cached_property
    def Purchases(self) -> zeep.client.Factory:
        return self._type_factory("purchases", "transactions")

    @cached_property
    def PurchasesTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.purchases", "transactions")

    @cached_property
    def Customers(self) -> zeep.client.Factory:
        return self._type_factory("customers", "transactions")

    @cached_property
    def CustomersTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.customers", "transactions")

    @cached_property
    def Financial(self) -> zeep.client.Factory:
        return self._type_factory("financial", "transactions")

    @cached_property
    def FinancialTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.financial", "transactions")

    @cached_property
    def Bank(self) -> zeep.client.Factory:
        return self._type_factory("bank", "transactions")

    @cached_property
    def BankTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.bank", "transactions")

    @cached_property
    def Inventory(self) -> zeep.client.Factory:
        return self._type_factory("inventory", "transactions")

    @cached_property
    def InventoryTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.inventory", "transactions")

    @cached_property
    def General(self) -> zeep.client.Factory:
        return self._type_factory("general", "transactions")

    @cached_property
    def Customization(self) -> zeep.client.Factory:
        return self._type_factory("customization", "setup")

    @cached_property
    def CustomizationTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.customization", "setup")

    @cached_property
    def Employees(self) -> zeep.client.Factory:
        return self._type_factory("employees", "lists")

    @cached_property
    def EmployeesTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.employees", "lists")

    @cached_property
    def Website(self) -> zeep.client.Factory:
        return self._type_factory("website", "lists")

    @cached_property
    def WebsiteTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.website", "lists")

    @cached_property
    def EmployeesTransactions(self) -> zeep.client.Factory:
        return self._type_factory("employees", "transactions")

    @cached_property
    def EmployeesTransactionsTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.employees", "transactions")

    @cached_property
    def Marketing(self) -> zeep.client.Factory:
        return self._type_factory("marketing", "lists")

    @cached_property
    def MarketingTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.marketing", "lists")

    @cached_property
    def DemandPlanning(self) -> zeep.client.Factory:
        return self._type_factory("demandplanning", "transactions")

    @cached_property
    def DemandPlanningTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.demandplanning", "transactions")

    @cached_property
    def SupplyChain(self) -> zeep.client.Factory:
        return self._type_factory("supplychain", "lists")

    @cached_property
    def SupplyChainTypes(self) -> zeep.client.Factory:
        return self._type_factory("types.supplychain", "lists")

    async def request(
        self, service_name: str, *args, additionalHeaders: Optional[dict] = None, **kw
    ):
        """
        Make a web service request to NetSuite

        Args:
            service_name:
                The NetSuite service to call
        Returns:
            The response from NetSuite
        """
        svc = getattr(self.service, service_name)
        headers = self.generate_passport()
        if additionalHeaders:
            headers.update(additionalHeaders)
        return await svc(*args, _soapheaders=headers, **kw)

    @WebServiceCall(
        "body.readResponseList.readResponse",
        extract=lambda resp: [r["record"] for r in resp],
    )
    async def getList(
        self,
        recordType: str,
        *,
        internalIds: Optional[Sequence[int]] = None,
        externalIds: Optional[Sequence[str]] = None,
    ) -> List[zeep.xsd.CompoundValue]:
        """Get a list of records"""
        if internalIds is None:
            internalIds = []
        else:
            internalIds = list(internalIds)
        if externalIds is None:
            externalIds = []
        else:
            externalIds = list(externalIds)

        if len(internalIds) + len(externalIds) == 0:
            return []

        return await self.request(
            "getList",
            self.Messages.GetListRequest(
                baseRef=[
                    self.Core.RecordRef(
                        type=recordType,
                        internalId=internalId,
                    )
                    for internalId in internalIds
                ]
                + [
                    self.Core.RecordRef(
                        type=recordType,
                        externalId=externalId,
                    )
                    for externalId in externalIds
                ],
            ),
        )

    @WebServiceCall(
        "body.readResponse",
        extract=lambda resp: resp["record"],
    )
    async def get(
        self,
        recordType: str,
        *,
        internalId: Optional[int] = None,
        externalId: Optional[str] = None,
    ) -> zeep.xsd.CompoundValue:
        """Get a single record"""
        if len([v for v in (internalId, externalId) if v is not None]) != 1:
            raise ValueError("Specify either `internalId` or `externalId`")

        if internalId:
            record_ref = self.Core.RecordRef(
                type=recordType,
                internalId=internalId,
            )
        else:
            record_ref = self.Core.RecordRef(
                type=recordType,
                externalId=externalId,
            )

        return await self.request("get", baseRef=record_ref)

    @WebServiceCall(
        "body.getAllResult",
        extract=lambda resp: resp["recordList"]["record"],
    )
    async def getAll(self, recordType: str) -> List[zeep.xsd.CompoundValue]:
        """Get all records of a given type."""
        return await self.request(
            "getAll",
            record=self.Core.GetAllRecord(
                recordType=recordType,
            ),
        )

    @WebServiceCall(
        "body.writeResponse",
        extract=lambda resp: resp["baseRef"],
    )
    async def add(self, record: zeep.xsd.CompoundValue) -> zeep.xsd.CompoundValue:
        """Insert a single record."""
        return await self.request("add", record=record)

    @WebServiceCall(
        "body.writeResponse",
        extract=lambda resp: resp["baseRef"],
    )
    async def update(self, record: zeep.xsd.CompoundValue) -> zeep.xsd.CompoundValue:
        """Insert a single record."""
        return await self.request("update", record=record)

    @WebServiceCall(
        "body.writeResponse",
        extract=lambda resp: resp["baseRef"],
    )
    async def upsert(self, record: zeep.xsd.CompoundValue) -> zeep.xsd.CompoundValue:
        """Upsert a single record."""
        return await self.request("upsert", record=record)

    @WebServiceCall(
        "body.searchResult",
        extract=lambda resp: resp["recordList"]["record"],
    )
    async def search(
        self, record: zeep.xsd.CompoundValue, additionalHeaders: Optional[dict] = None
    ) -> List[zeep.xsd.CompoundValue]:
        """Search records"""
        return await self.request(
            "search", searchRecord=record, additionalHeaders=additionalHeaders
        )

    @WebServiceCall(
        "body.searchResult",
        extract=lambda resp: resp["recordList"]["record"],
    )
    async def searchMoreWithId(
        self, searchId: str, pageIndex: int
    ) -> List[zeep.xsd.CompoundValue]:
        """Search records with pagination"""
        return await self.request(
            "searchMoreWithId", searchId=searchId, pageIndex=pageIndex
        )

    @WebServiceCall(
        "body.writeResponseList",
        extract=lambda resp: [record["baseRef"] for record in resp],
    )
    async def upsertList(
        self, records: List[zeep.xsd.CompoundValue]
    ) -> List[zeep.xsd.CompoundValue]:
        """Upsert a list of records."""
        return await self.request("upsertList", record=records)

    @WebServiceCall(
        "body.getItemAvailabilityResult",
        extract=lambda resp: resp["itemAvailabilityList"]["itemAvailability"],
        default=[],
    )
    async def getItemAvailability(
        self,
        *,
        internalIds: Optional[Sequence[int]] = None,
        externalIds: Optional[Sequence[str]] = None,
        lastQtyAvailableChange: Optional[datetime] = None,
    ) -> List[Dict]:
        if internalIds is None:
            internalIds = []
        else:
            internalIds = list(internalIds)
        if externalIds is None:
            externalIds = []
        else:
            externalIds = list(externalIds)

        if len(internalIds) + len(externalIds) == 0:
            return []

        item_filters = [
            {"type": "inventoryItem", "internalId": internalId}
            for internalId in internalIds
        ] + [
            {"type": "inventoryItem", "externalId": externalId}
            for externalId in externalIds
        ]

        return await self.request(
            "getItemAvailability",
            itemAvailabilityFilter=[
                {
                    "item": {"recordRef": item_filters},
                    "lastQtyAvailableChange": lastQtyAvailableChange,
                }
            ],
        )



import base64
import hmac
import random
from datetime import datetime
from typing import Dict, TypeVar

from ..config import Config, TokenAuth

NetSuite = TypeVar("NetSuite")


class Passport:
    def get_element(self) -> str:
        raise NotImplementedError


class TokenPassport(Passport):
    def __init__(
        self,
        ns: NetSuite,
        *,
        account: str,
        consumer_key: str,
        consumer_secret: str,
        token_id: str,
        token_secret: str,
    ) -> None:
        self.ns = ns
        self.account = account
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.token_id = token_id
        self.token_secret = token_secret

    def _generate_timestamp(self) -> str:
        """Generate timestamp

        Returns:
            str: A seconds precision timestamp
        """
        return str(int(datetime.now().timestamp()))

    def _generate_nonce(self, length: int = 20) -> str:
        """Generate pseudorandom number"""
        return "".join([str(random.randint(0, 9)) for i in range(length)])

    def _get_signature_message(self, nonce: str, timestamp: str) -> str:
        return "&".join(
            (
                self.account,
                self.consumer_key,
                self.token_id,
                nonce,
                timestamp,
            )
        )

    def _get_signature_key(self) -> str:
        return "&".join((self.consumer_secret, self.token_secret))

    def _get_signature_value(self, nonce: str, timestamp: str) -> str:
        key = self._get_signature_key()
        message = self._get_signature_message(nonce, timestamp)
        hashed = hmac.new(
            key=key.encode("utf-8"), msg=message.encode("utf-8"), digestmod="sha256"
        ).digest()
        return base64.b64encode(hashed).decode()

    def _get_signature(self, nonce: str, timestamp: str):
        return self.ns.Core.TokenPassportSignature(  # type: ignore[attr-defined]
            self._get_signature_value(nonce, timestamp),
            algorithm="HMAC-SHA256",
        )

    def get_element(self):
        nonce = self._generate_nonce()
        timestamp = self._generate_timestamp()
        signature = self._get_signature(nonce, timestamp)
        return self.ns.Core.TokenPassport(
            account=self.account,
            consumerKey=self.consumer_key,
            token=self.token_id,
            nonce=nonce,
            timestamp=timestamp,
            signature=signature,
        )


def make(ns: NetSuite, config: Config) -> Dict:
    auth = config.auth
    if isinstance(auth, TokenAuth):
        token_passport = TokenPassport(
            ns,
            account=config.account,
            consumer_key=auth.consumer_key,
            consumer_secret=auth.consumer_secret,
            token_id=auth.token_id,
            token_secret=auth.token_secret,
        )
        return {"tokenPassport": token_passport.get_element()}
    else:
        raise NotImplementedError(auth.__class__)



import asyncio
import logging
from functools import cached_property

import httpx
from authlib.integrations.httpx_client import OAuth1Auth
from authlib.oauth1.rfc5849.client_auth import ClientAuth
from authlib.oauth1.rfc5849.signature import generate_signature_base_string
from oauthlib.oauth1.rfc5849.signature import sign_hmac_sha256

from . import json
from .exceptions import NetsuiteAPIRequestError, NetsuiteAPIResponseParsingError

__all__ = ("RestApiBase",)

DEFAULT_SIGNATURE_METHOD = "HMAC-SHA256"

logger = logging.getLogger(__name__)


def authlib_hmac_sha256_sign_method(client, request):
    """Sign a HMAC-SHA256 signature."""
    base_string = generate_signature_base_string(request)
    return sign_hmac_sha256(base_string, client.client_secret, client.token_secret)


ClientAuth.register_signature_method("HMAC-SHA256", authlib_hmac_sha256_sign_method)


class RestApiBase:
    _concurrent_requests: int = 10
    _default_timeout: int = 10
    _signature_method: str = DEFAULT_SIGNATURE_METHOD

    @cached_property
    def _request_semaphore(self) -> asyncio.Semaphore:
        # NOTE: Shouldn't be put in __init__ as we might not have a running
        #       event loop at that time.
        return asyncio.Semaphore(self._concurrent_requests)

    async def _request(self, method: str, subpath: str, **request_kw):
        resp = await self._request_impl(method, subpath, **request_kw)

        if resp.status_code < 200 or resp.status_code > 299:
            raise NetsuiteAPIRequestError(resp.status_code, resp.text)

        if resp.status_code == 204:
            return None
        else:
            try:
                return json.loads(resp.text)
            except Exception:
                raise NetsuiteAPIResponseParsingError(resp.status_code, resp.text)

    async def _request_impl(
        self, method: str, subpath: str, **request_kw
    ) -> httpx.Response:
        method = method.upper()
        url = request_kw.pop("url", self._make_url(subpath))

        headers = {**self._make_default_headers(), **request_kw.pop("headers", {})}

        timeout = request_kw.pop("timeout", self._default_timeout)

        if "json" in request_kw:
            request_kw["data"] = json.dumps(request_kw.pop("json"))

        kw = {**request_kw}
        logger.debug(
            f"Making {method.upper()} request to {url}. Keyword arguments: {kw}"
        )

        async with self._request_semaphore:
            async with httpx.AsyncClient() as c:
                resp = await c.request(
                    method=method,
                    url=url,
                    headers=headers,
                    auth=self._make_auth(),
                    timeout=timeout,
                    **kw,
                )

        resp_headers_json = json.dumps(dict(resp.headers))
        logger.debug(f"Got response headers from NetSuite: {resp_headers_json}")

        return resp

    def _make_url(self, subpath: str):
        raise NotImplementedError

    def _make_auth(self):
        auth = self._config.auth
        return OAuth1Auth(
            client_id=auth.consumer_key,
            client_secret=auth.consumer_secret,
            token=auth.token_id,
            token_secret=auth.token_secret,
            realm=self._config.account,
            force_include_body=True,
            signature_method=self._signature_method,
        )

    def _make_default_headers(self):
        return {"Content-Type": "application/json"}



import argparse
import asyncio
import inspect
import logging
import sys

from ..config import Config
from ..constants import DEFAULT_INI_PATH, DEFAULT_INI_SECTION
from . import helpers, interact, misc, rest_api, restlet, soap_api

# include pretty_traceback if it exists for better dev tracebacks in CLI
try:
    import pretty_traceback

    pretty_traceback.install()
except ImportError:
    pass

__all__ = ("main",)


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "-l",
        "--log-level",
        help="The log level to use",
        default=None,
        choices=("DEBUG", "INFO", "WARNING", "ERROR"),
    )
    parser.add_argument(
        "-p",
        "--config-path",
        help="The config file to get settings from",
        default=DEFAULT_INI_PATH,
    )
    parser.add_argument(
        "-c",
        "--config-section",
        help="The config section to get settings from",
        default=DEFAULT_INI_SECTION,
    )

    parser.add_argument(
        "--config-environment",
        help="Use environment variables for configuration",
        action="store_true",
    )

    subparser = parser.add_subparsers(help="App CLI", required=True)

    misc.add_parser(parser, subparser)
    interact.add_parser(parser, subparser)
    restlet_parser, _ = restlet.add_parser(parser, subparser)
    rest_api_parser, _ = rest_api.add_parser(parser, subparser)
    soap_api_parser, _ = soap_api.add_parser(parser, subparser)

    try:
        args = parser.parse_args()
    except Exception:
        parser.print_help()
        return

    subparser_name = sys.argv[-1]

    # Call version directly to avoid loading of config
    if subparser_name == "version":
        print(args.func())
        return

    # Show help section instead of an error when no arguments were passed...
    if subparser_name == "rest-api":
        rest_api_parser.print_help()
        return
    elif subparser_name == "soap-api":
        rest_api_parser.print_help()
        return
    elif subparser_name == "restlet":
        restlet_parser.print_help()
        return

    config = None

    if args.config_environment:
        config = Config.from_env()
    else:
        config = helpers.load_config_or_error(
            parser, args.config_path, args.config_section
        )

    log_level = args.log_level

    if log_level is None:
        log_level = config.log_level

    log_level_number = getattr(logging, log_level)
    logging.basicConfig(level=log_level_number)

    ret = args.func(config, args)

    if inspect.iscoroutinefunction(args.func):
        ret = asyncio.run(ret)

    if ret is not None:
        print(ret)



from functools import cached_property
from typing import Any, Dict, Optional

from .config import Config
from .rest_api import NetSuiteRestApi
from .restlet import NetSuiteRestlet
from .soap_api import NetSuiteSoapApi

__all__ = ("NetSuite",)


class NetSuite:
    def __init__(
        self,
        config: Config,
        *,
        soap_api_options: Optional[Dict[str, Any]] = None,
        rest_api_options: Optional[Dict[str, Any]] = None,
        restlet_options: Optional[Dict[str, Any]] = None,
    ):
        self._config = config
        self._soap_api_options = soap_api_options or {}
        self._rest_api_options = rest_api_options or {}
        self._restlet_options = restlet_options or {}

    @cached_property
    def rest_api(self) -> NetSuiteRestApi:
        return NetSuiteRestApi(self._config, **self._rest_api_options)

    @cached_property
    def soap_api(self) -> NetSuiteSoapApi:
        return NetSuiteSoapApi(self._config, **self._soap_api_options)

    @cached_property
    def restlet(self) -> NetSuiteRestlet:
        return NetSuiteRestlet(self._config, **self._restlet_options)



from .. import json
from ..client import NetSuite
from ..config import Config
from ..soap_api import helpers

__all__ = ()


def add_parser(parser, subparser):
    soap_api_parser = subparser.add_parser(
        "soap-api", description="Make NetSuite SuiteTalk Web Services SOAP requests"
    )
    soap_api_subparser = soap_api_parser.add_subparsers()
    _add_get_parser(soap_api_parser, soap_api_subparser)
    _add_get_list_parser(soap_api_parser, soap_api_subparser)

    return (soap_api_parser, soap_api_subparser)


def _add_get_list_parser(parser, subparser):
    async def getList(config, args) -> str:
        soap_api = _get_soap_api_or_error(parser, config)
        resp = await soap_api.getList(
            args.record_type, externalIds=args.externalId, internalIds=args.internalId
        )
        return _dump_response(resp)

    p = subparser.add_parser("getList", description="Call the getList method")
    p.add_argument("record_type", help="The record type to get")
    p.add_argument(
        "-e",
        "--externalId",
        action="append",
        help="External IDs to get",
    )
    p.add_argument(
        "-i",
        "--internalId",
        action="append",
        help="Internal IDs to get",
    )
    p.set_defaults(func=getList)


def _add_get_parser(parser, subparser):
    async def get(config, args) -> str:
        soap_api = _get_soap_api_or_error(parser, config)
        resp = await soap_api.get(
            args.record_type, externalId=args.externalId, internalId=args.internalId
        )
        return _dump_response(resp)

    p = subparser.add_parser("get", description="Call the `get` method")
    p.add_argument("record_type", help="The record type to get")
    p.add_argument(
        "-e",
        "--externalId",
        help="External ID to get",
    )
    p.add_argument(
        "-i",
        "--internalId",
        help="Internal ID to get",
    )
    p.set_defaults(func=get)


def _get_soap_api_or_error(parser, config: Config):
    ns = NetSuite(config)

    try:
        return ns.soap_api  # Cached property that initializes NetSuiteRestApi
    except RuntimeError as ex:
        parser.error(str(ex))


def _dump_response(resp) -> str:
    return json.dumps(helpers.to_builtin(resp))



import asyncio

import IPython
import traitlets

from ..client import NetSuite

__all__ = ()


def add_parser(parser, subparser):
    interact_parser = subparser.add_parser(
        "interact",
        description="Starts a REPL to enable live interaction with NetSuite webservices",
    )
    interact_parser.set_defaults(func=interact)
    return (interact_parser, None)


def interact(config, args):
    ns = NetSuite(config)

    user_ns = {
        "ns": ns,
        "query": lambda **kwargs: asyncio.run(ns.rest_api.suiteql(**kwargs)),
    }

    banner1 = """Welcome to Netsuite WS client interactive mode
Available vars:
    `ns` - NetSuite client
    `query` - run SuiteQL sync

Example usage:
    soap_results = await ns.soap_api.getList('customer', internalIds=[1337])

    restlet_results = await ns.restlet.get(987, deploy=2)

    rest_api_results = await ns.rest_api.get("/record/v1/salesOrder")
    ns.rest_api._default_timeout = 60 * 5
    catalog = await ns.rest_api.openapi()

    query_results = await ns.rest_api.suiteql("SELECT * FROM salesOrder")
    query_results = query("SELECT * FROM salesOrder")
"""

    print(banner1)
    IPython.start_ipython(argv=[], user_ns=user_ns)



import logging
from functools import cached_property

from . import rest_api_base
from .config import Config

logger = logging.getLogger(__name__)

__all__ = ("NetSuiteRestlet",)


class NetSuiteRestlet(rest_api_base.RestApiBase):
    def __init__(
        self,
        config: Config,
        *,
        default_timeout: int = 60,
        concurrent_requests: int = 10,
        signature_method: str = rest_api_base.DEFAULT_SIGNATURE_METHOD,
    ):
        self._config = config
        self._default_timeout = default_timeout
        self._concurrent_requests = concurrent_requests
        self._signature_method = signature_method

    @cached_property
    def hostname(self) -> str:
        return self._make_hostname()

    async def get(self, script_id: int, *, deploy: int = 1, **request_kw):
        subpath = self._make_restlet_params(script_id, deploy)
        return await self._request("GET", subpath, **request_kw)

    async def post(self, script_id: int, *, deploy: int = 1, **request_kw):
        subpath = self._make_restlet_params(script_id, deploy)
        return await self._request("POST", subpath, **request_kw)

    async def put(self, script_id: int, *, deploy: int = 1, **request_kw):
        subpath = self._make_restlet_params(script_id, deploy)
        return await self._request("PUT", subpath, **request_kw)

    async def delete(self, script_id: int, *, deploy: int = 1, **request_kw):
        subpath = self._make_restlet_params(script_id, deploy)
        return await self._request("DELETE", subpath, **request_kw)

    def _make_restlet_params(self, script_id: int, deploy: int = 1) -> str:
        return f"?script={script_id}&deploy={deploy}"

    def _make_hostname(self):
        return f"{self._config.account_slugified}.restlets.api.netsuite.com"

    def _make_url(self, subpath: str) -> str:
        return f"https://{self.hostname}/app/site/hosting/restlet.nl{subpath}"
